# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'gio-pyio'
copyright = '2024, Christoph Matthias Kohnen'
author = 'Christoph Matthias Kohnen'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
exclude_patterns = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'pygobject': ('https://amolenaar.pages.gitlab.gnome.org/pygobject-docs/', None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



"""gio_pyio lib."""
import io
import os

from gi.repository import GLib, Gio


class StreamWrapper(io.IOBase):
    """Wrap a stream as a `file object`_.

    See :func:`Gio.open_file_like` for a convenience method to open a file as a
    `file object`_. Note, that this does not implement buffering, seeking, etc.
    and relies on the capabilities of *stream*.

    :param stream stream:
        A stream to be wrapped.
    :raises TypeError:
        Invalid argument.

    .. _file object: https://docs.python.org/3/glossary.html#term-file-object
    """

    def __init__(self, stream):
        if isinstance(stream, Gio.InputStream):
            self._input_stream = stream
            self._output_stream = None
            self._ref_stream = stream
        elif isinstance(stream, Gio.OutputStream):
            self._input_stream = None
            self._output_stream = stream
            self._ref_stream = stream
        elif isinstance(stream, Gio.IOStream):
            # For some methods, we assume, both stream represent the same
            # object as well as being in sync in regards to seeking, that way
            # we don't need to duplicate logic for most methods.
            self._input_stream = stream.get_input_stream()
            self._output_stream = stream.get_output_stream()
            self._ref_stream = self._input_stream
            # Keep a reference, or the stream might get closed
            self._io_stream = stream
        else:
            raise TypeError('expected stream, got %s' % type(stream))

    def close(self):
        """Flush and close the underlying stream.

        This method has no effect if the underlying stream is already closed.
        Once closed, any operation (e. g. reading or writing) will raise a
        ValueError.

        As a convenience, it is allowed to call this method more than once;
        only the first call, however, will have an effect.
        """
        if hasattr(self, '_io_stream'):
            self._io_stream.close()
            return
        if self.readable():
            self._input_stream.close()
        if self.writable():
            self._output_stream.close()

    @property
    def closed(self):
        """``True`` if the underlying stream is closed."""
        return self._ref_stream.is_closed()

    def fileno(self):
        """Return the underlying file descriptor if it exists.

        :rtype: int
        :returns:
            The underlying file descriptor.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream is not based on a file descriptor.
        """
        self._checkClosed()
        if not (self._ref_stream, 'get_fd'):
            self._unsupported('fileno')
        return self._ref_stream.get_fd()

    def flush(self):
        """Flush the write buffers of the underlying stream if applicable.

        This does nothing for read-only streams.

        :raises ValueError:
            If the underlying stream is closed.
        """
        self._checkClosed()
        if self.writable():
            self._output_stream.flush(None)

    def readable(self):
        """Whether or not the stream is readable.

        :rtype bool:
        :returns:
            Whether or not this wrapper can be read from.
        """
        return self._input_stream is not None

    def read(self, size=-1):
        """Read up to *size* bytes from the underlying stream and return them.

        As a convenience if *size* is unspecified or -1, all bytes until EOF
        are returned. The result may be fewer bytes than requested, if EOF is
        reached.

        :param int size:
            The amount of bytes to read from the underlying stream.
        :rtype: bytes
        :returns:
            Bytes read from the underlying stream.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream is not readable.
        """
        self._checkClosed()
        self._checkReadable()
        if size == 0:
            return b''
        elif size > 0:
            return self._input_stream.read_bytes(size, None).get_data()
        # Try to determine the length of the stream, else fall back on
        # default buffer size for reading
        if isinstance(self._input_stream, Gio.BufferedInputStream):
            def_bufsize = self._input_stream.get_buffer_size()
        else:
            def_bufsize = io.DEFAULT_BUFFER_SIZE
            end = None
            if hasattr(self._input_stream, 'get_size'):
                end = self._input_stream.get_size()
            elif hasattr(self._input_stream, 'query_info'):
                info = self._input_stream.query_info('standard::size', None)
                end = info.get_size()
            if end is not None:
                pos = self._ref_stream.tell()
                if end >= pos:
                    bufsize = end - pos + 1
            else:
                bufsize = def_bufsize
        result = bytearray()
        while True:
            if len(result) >= bufsize:
                bufsize = len(result)
                bufsize += max(bufsize, def_bufsize)
            n = bufsize - len(result)
            chunk = self._input_stream.read_bytes(n, None)
            if chunk.get_size() == 0:  # EOF reached
                break
            result += chunk.get_data()
        return bytes(result)

    read1 = read
    readall = read

    def readinto(self, b):
        """Read bytes into a pre-allocated, writable `bytes-like object`_ *b*.

        :param bytes-like b:
            A pre-allocated object.
        :rtype: int
        :returns:
            Number of bytes written.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream is not readable.

        .. _bytes-like object: https://docs.python.org/3/glossary.html#term-bytes-like-object
        """
        self._checkClosed()
        self._checkReadable()
        view = memoryview(b).cast('B')
        data = self._input_stream.read_bytes(len(view), None)
        size = data.get_size()
        view[:size] = data.get_data()
        return size

    readinto1 = readinto

    def seek(self, offset, whence=os.SEEK_SET):
        """Change the underlying stream position.

        *offset* is interpreted relative to the position indicated by *whence*.

        :param int offset:
            Where to change the stream position to, relative to *whence*
        :param int whence:
            Reference for *offset*. Values are:
            * 0 -- start of stream (the default); offset should'nt be negative
            * 1 -- current stream position; offset may be negative
            * 2 -- end of stream; offset is usually negative
        :rtype: int
        :returns:
            The new absolute position of the underlying stream.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream is not seekable.
        """
        self._checkClosed()
        self._checkSeekable()
        # Enum values in python and Gio:
        # 0: os.SEEK_SET : Gio.SeekType.CUR
        # 1: os.SEEK_CUR : Gio.SeekType.SET
        # 2: os.SEEK_END : Gio.SeekType.END
        # so 1 and 0 need to be switched
        if whence != 2:
            whence = not whence
        if self.readable():
            self._input_stream.seek(offset, whence, None)
        if self.writable():
            self._output_stream.seek(offset, whence, None)
        return self._ref_stream.tell()

    def seekable(self):
        """
        Whether or not the stream is seekable.

        :rtype: bool
        :returns:
            Whether or not the underlying stream supports seeking.
        :raises ValueError:
            If the underlying stream is closed.
        """
        self._checkClosed()
        return self._ref_stream.can_seek()

    def tell(self):
        """
        Tell the current stream position.

        :rtype: int
        :returns:
            The position of the underlying stream.
        :raises ValueError:
            If the underlying stream is closed.
        """
        self._checkClosed()
        return self._ref_stream.tell()

    def truncate(self, size=None):
        """Resize the underlying stream to *size*.

        :param int size:
            The size, the stream should be set to. If ``None`` the current
            position is used.
        :rtype: int
        :returns:
            The new size of the underlying stream.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream can not be written to.
        """
        self._checkClosed()
        if self._output_stream is None or \
                not self._output_stream.can_truncate():
            raise io.UnsupportedOperation('truncate')
        if size is None:
            size = self._output_stream.tell()
        self._output_stream.truncate(size)
        return size

    def writable(self):
        """
        Wheter or not the stream can be written to.

        :rtype: bool
        :returns:
            Whether or not this wrapper can be written to.
        """
        return self._output_stream is not None

    def write(self, b):
        """Write *b* to the underlying stream.

        :param bytes-like b:
            Content to be written to the underlying stream.
        :rtype: int
        :returns:
            The number of bytes written to the underlying stream.
        :raises ValueError:
            If the underlying stream is closed.
        :raises io.UnsupportedOperationException:
            If the underlying stream can not be written to.
        """
        self._checkClosed()
        self._checkWritable()
        if b is None or b == b'':
            return 0
        return self._output_stream.write_bytes(GLib.Bytes(b))


def open(file, mode='r', buffering=-1, encoding=None, errors=None,
         newline=None, native=True):
    r"""Open the file and create a corresponding `file object`_.

    If the file cannot be opened, an OSError is raised. This behaves analog to
    pythons builtin :external:py:func:`open` function. See
    `Reading and Writing Files`_ for examples of io using python.

    :param Gio.File file:
        The file to open.
    :param str mode:
        Mode in which the file is opened. Defaults to 'r' which means
        open for reading in text mode. Other common values are 'w' for
        writing (truncating the file if it already exists), 'x' for
        exclusive creation of a new file, and 'a' for appending. The
        available modes are:

        ========= ===================================================
        Character Meaning
        --------- ---------------------------------------------------
        'r'       open for reading (default)
        'w'       open for writing, truncating the file first
        'x'       create a new file and open it for writing
        'a'       open for writing, appending to the end of the file
        'b'       binary mode
        't'       text mode (default)
        '+'       open a disk file for updating (reading and writing)
        ========= ===================================================

        The default value is 'r' (open for reading text, a synonym of 'rt').
        For binary random access, the mode 'w+b' opens and truncates the
        file to 0 bytes, while 'r+b' opens the file without truncation. The
        'x' mode implies 'w' and raises an `FileExistsError` if the file
        already exists.

        Python distinguishes between files opened in binary and text mode.
        Files opened in binary mode (appending 'b' to the mode argument)
        return contents as bytes objects without any decoding. In text mode
        (the default, or when 't' is appended to the mode argument), the
        contents of the file are returned as strings, the bytes having
        first been decoded using *encoding*.
    :param int buffering:
        Buffering policy. Pass 0 to switch buffering off (only allowed in
        binary mode), 1 to select line buffering (only usable in text mode),
        and an integer > 1 to indicate the size of a fixed-size chunk
        buffer. When no buffering argument is given.

        Files are buffered in fixed-size chunks; the size of the buffer is
        chosen using a heuristic trying to determine the underlying device's
        blksize and falling back on `io.DEFAULT_BUFFER_SIZE`. The buffer
        will typically be 4096 or 8192 bytes long.
    :param str encoding:
        Encoding used to decode or encode the file. Can only be used in text
        mode. The default encoding is platform dependent, but any encoding
        supported by Python can be passed. See the codecs module for the
        list of supported encodings.
    :param str errors:
        How encoding errors are to be handled. Can not be used in binary
        mode. Pass 'strict' to raise a ValueError exception if there is an
        encoding error (the default of None has the same effect), or pass
        'ignore' to ignore errors. (Note that ignoring encoding errors can
        lead to data loss.) See the documentation for codecs.register for a
        list of the permitted encoding error strings.
    :param str newline:
        How universal newlines work (only applies to text mode). Can be
        None, '', '\n', '\r', and '\r\n'. Works as follows:

        * On input, if newline is None, universal newlines mode is enabled.
          Lines in the input can end in '\n', '\r', or '\r\n', and these are
          translated into '\n' before being returned to the caller. If it is
          '', universal newline mode is enabled, but line endings are
          returned to the caller untranslated. If it has any of the other
          legal values, input lines are only terminated by the given string,
          and the line ending is returned to the caller untranslated.

        * On output, if newline is None, any '\n' characters written are
          translated to the system default line separator, os.linesep. If
          newline is '', no translation takes place. If newline is any of
          the other legal values, any '\n' characters written are translated
          to the given string.
    :param bool native:
        Try and obtain a file descriptor and use python standard io libraries.
        If False, the result will always be a wrapped Gio stream.
    :rtype: file-like
    :returns:
        A new `file object`_. When used to open a file in a text mode ('w',
        'r', 'wt', 'rt', etc.), the object will be a TextIOWrapper. When
        used to open a file in a binary mode, the returned class varies:
        in read binary mode, it will be a BufferedReader; in write binary
        and append binary modes, it will be a BufferedWriter, and in
        read/write mode, it will be a BufferedRandom. If buffering is
        disabled, the object will either be a FileIO or
        :py:class:`StreamWrapper` depending on python native libraries can be
        used.
    :raises TypeError:
        Invalid argument passed.
    :raises ValueError:
        Invalid argument passed.
    :raises OSError:
        Failed to open file.

    .. _file object: https://docs.python.org/3/glossary.html#term-file-object
    .. _Reading and Writing Files: https://docs.python.org/3/tutorial/inputoutput.html#tut-files
    """
    # Argument validation
    if not isinstance(mode, str):
        raise TypeError('invalid mode: %r' % mode)
    if not isinstance(buffering, int):
        raise TypeError('invalid buffering: %r' % buffering)
    modes = set(mode)
    if modes - set('axrwb+t') or len(mode) > len(modes):
        raise ValueError('invalid mode: %r' % mode)
    if encoding is not None and not isinstance(encoding, str):
        raise TypeError('invalid encoding: %r' % encoding)
    if errors is not None and not isinstance(errors, str):
        raise TypeError('invalid errors: %r' % errors)
    creating = 'x' in modes
    reading = 'r' in modes
    writing = 'w' in modes
    appending = 'a' in modes
    updating = '+' in modes
    binary = 'b' in modes
    if binary and 't' in modes:
        raise ValueError("can't have text and binary mode at once")
    if creating + reading + writing + appending > 1:
        raise ValueError('must have exactly one of create/read/write/append'
                         ' mode')
    if not (creating or reading or writing or appending):
        raise ValueError('Must have exactly one of create/read/write/append'
                         ' mode and at most one plus')
    if binary and encoding is not None:
        raise ValueError("binary mode doesn't take an encoding argument")
    if binary and errors is not None:
        raise ValueError("binary mode doesn't take an errors argument")
    if binary and newline is not None:
        raise ValueError("binary mode doesn't take a newline argument")

    # For non-native files we use the result of `file.get_basename()`
    rep_str = file.peek_path() if file.is_native() else file.get_basename()
    file_type = file.query_file_type(Gio.FileQueryInfoFlags.NONE, None)
    if file_type == Gio.FileType.DIRECTORY:
        raise OSError(21, "Is a directory: '%s'" % rep_str)
    if file.query_exists():
        if creating:
            raise OSError(17, "File exists: '%s'" % rep_str)
    elif reading:
        raise OSError(2, "No such file or directory: '%s'" % rep_str)
    if buffering == 0 and not binary:
        raise ValueError("can't have unbuffered text I/O")

    if native and file.is_native():
        file_like = io.FileIO(
            file.peek_path(),
            (creating and 'x' or '') +
            (reading and 'r' or '') +
            (writing and 'w' or '') +
            (appending and 'a' or '') +
            (updating and '+' or ''),
        )
    else:
        stream = None
        # Match given mode to respective opener. All calls are non-async, thus
        # blocking as well as not cancellable.
        if updating:
            if creating:
                stream = file.create_readwrite(Gio.FileCreateFlags.NONE, None)
            elif writing:
                stream = file.replace_readwrite(None, False,
                                                Gio.FileCreateFlags.NONE, None)
            else:
                stream = file.open_readwrite(None)
        else:
            if creating:
                stream = file.create(Gio.FileCreateFlags.NONE, None)
            elif reading:
                stream = file.read(None)
            elif writing:
                stream = file.replace(None, False, Gio.FileCreateFlags.NONE,
                                      None)
            elif appending:
                stream = file.append_to(Gio.FileCreateFlags.NONE, None)

        # at this point stream should not be `None` or input validation has
        # failed substantially
        assert stream is not None
        file_like = StreamWrapper(stream)
    line_buffering = False
    if buffering != 0:
        if buffering == 1:
            buffering = -1
            line_buffering = True
        if buffering < 0:
            buffering = io.DEFAULT_BUFFER_SIZE
            try:
                # Try to set buffersize to the blksize of the file system
                blksize = os.fstat(file_like.fileno()).st_blksize
                if blksize > 1:
                    buffering = blksize
            except (OSError, AttributeError):
                pass
        if updating:
            wrapper = io.BufferedRandom
        elif reading:
            wrapper = io.BufferedReader
        else:
            wrapper = io.BufferedWriter
        file_like = wrapper(file_like, buffering)
    if not binary:
        file_like = io.TextIOWrapper(file_like, encoding=encoding,
                                     errors=errors, newline=newline,
                                     line_buffering=line_buffering)
        file_like.mode = mode
    return file_like



# These test cases are adapted from python's own test for file validation
# They are licensed under the
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# https://github.com/python/cpython/blob/main/LICENSE

import contextlib
import gc
import json
import subprocess
import sys
import unittest
from array import array
from collections import UserList
from pathlib import Path
from weakref import proxy

from gi.repository import GLib, Gio

import gio_pyio


class GioFileLikeTests(unittest.TestCase):

    def setUp(self):
        self.file, stream = Gio.File.new_tmp('TestGFile.XXXXXX')
        stream.close()
        self.f = gio_pyio.open(self.file, 'wb', native=False)

    def tearDown(self):
        if self.f:
            self.f.close()
        with contextlib.suppress(GLib.Error):
            self.file.delete(None)

    def testWeakRefs(self):
        # verify weak references
        p = proxy(self.f)
        p.write(b'teststring')
        self.assertEqual(self.f.tell(), p.tell())
        self.f.close()
        self.f = None
        gc.collect()
        self.assertRaises(ReferenceError, getattr, p, 'tell')

    def testSeekTell(self):
        self.f.write(bytes(range(20)))
        self.assertEqual(self.f.tell(), 20)
        self.f.seek(0)
        self.assertEqual(self.f.tell(), 0)
        self.f.seek(10)
        self.assertEqual(self.f.tell(), 10)
        self.f.seek(5, 1)
        self.assertEqual(self.f.tell(), 15)
        self.f.seek(-5, 1)
        self.assertEqual(self.f.tell(), 10)
        self.f.seek(-5, 2)
        self.assertEqual(self.f.tell(), 15)

    def testAttributes(self):
        # verify expected attributes exist
        self.f.closed   # merely shouldn't blow up

    def testReadinto(self):
        # verify readinto
        self.f.write(b'12')
        self.f.close()
        a = array('b', b'x' * 10)
        self.f = gio_pyio.open(self.file, 'rb', native=False)
        n = self.f.readinto(a)
        self.assertEqual(b'12', a.tobytes()[:n])

    def testReadinto_text(self):
        # verify readinto refuses text files
        a = array('b', b'x' * 10)
        self.f.close()
        self.f = gio_pyio.open(self.file, encoding='utf-8', native=False)
        if hasattr(self.f, 'readinto'):
            self.assertRaises(TypeError, self.f.readinto, a)

    def testReadintoByteArray(self):
        self.f.write(bytes([1, 2, 0, 255]))
        self.f.close()

        ba = bytearray(b'abcdefgh')
        with gio_pyio.open(self.file, 'rb', native=False) as f:
            n = f.readinto(ba)
        self.assertEqual(ba, b'\x01\x02\x00\xffefgh')
        self.assertEqual(n, 4)

    def testWritelinesUserList(self):
        # verify writelines with instance sequence
        userlist = UserList([b'1', b'2'])
        self.f.writelines(userlist)
        self.f.close()
        self.f = gio_pyio.open(self.file, 'rb', native=False)
        buf = self.f.read()
        self.assertEqual(buf, b'12')

    def testWritelinesIntegers(self):
        # verify writelines with integers
        self.assertRaises(TypeError, self.f.writelines, [1, 2, 3])

    def testWritelinesIntegersUserList(self):
        # verify writelines with integers in UserList
        userlist = UserList([1, 2, 3])
        self.assertRaises(TypeError, self.f.writelines, userlist)

    def testWritelinesNonString(self):
        # verify writelines with non-string object
        class NonString:
            pass

        self.assertRaises(TypeError, self.f.writelines,
                          [NonString(), NonString()])

    def testWritelinesError(self):
        self.assertRaises(TypeError, self.f.writelines, [1, 2, 3])
        self.assertRaises(TypeError, self.f.writelines, None)
        self.assertRaises(TypeError, self.f.writelines, 'abc')

    def testErrors(self):
        f = self.f
        self.assertFalse(f.isatty())
        self.assertFalse(f.closed)

        if hasattr(f, 'readinto'):
            self.assertRaises((OSError, TypeError), f.readinto, '')
        f.close()
        self.assertTrue(f.closed)

    def testReject(self):
        self.assertRaises(TypeError, self.f.write, 'Hello!')

    def testMethods(self):
        methods = [('fileno', ()),
                   ('flush', ()),
                   ('isatty', ()),
                   ('__next__', ()),
                   ('read', ()),
                   ('write', (b'',)),
                   ('readline', ()),
                   ('readlines', ()),
                   ('seek', (0,)),
                   ('tell', ()),
                   ('write', (b'',)),
                   ('writelines', ([],)),
                   ('__iter__', ()),
                   ('truncate', ()),
                   ]

        # __exit__ should close the file
        self.f.__exit__(None, None, None)
        self.assertTrue(self.f.closed)

        for methodname, args in methods:
            method = getattr(self.f, methodname)
            # should raise on closed file
            self.assertRaises(ValueError, method, *args)

        # file is closed, __exit__ shouldn't do anything
        self.assertEqual(self.f.__exit__(None, None, None), None)
        # it must also return None if an exception was given
        try:
            1 / 0
        except ZeroDivisionError:
            self.assertEqual(self.f.__exit__(*sys.exc_info()), None)

    def testReadWhenWriting(self):
        self.assertRaises(OSError, self.f.read)

    def testModeStrings(self):
        # check invalid mode strings
        for mode in ('', 'aU', 'wU+', 'U+', '+U', 'rU+'):
            try:
                f = gio_pyio.open(self.file, mode)
            except ValueError:
                pass
            else:
                f.close()
                self.fail('%r is an invalid file mode' % mode)
        # check valid mode strings
        for mode in ('rt', 'wb', 'a+'):
            f = gio_pyio.open(self.file, mode, native=False)
            f.close()
            f = gio_pyio.open(self.file, mode, native=True)
            f.close()

    def testBadModeArgument(self):
        # verify that we get a sensible error message for bad mode argument
        bad_mode = 'qwerty'
        try:
            f = gio_pyio.open(self.file, bad_mode)
        except ValueError as msg:
            if msg.args[0] != 0:
                s = str(msg)
                if bad_mode not in s:
                    self.fail('bad error message for invalid mode: %s' % s)
            # if msg.args[0] == 0, we're probably on Windows where there may be
            # no obvious way to discover why open() failed.
        else:
            f.close()
            self.fail('no error for invalid mode: %s' % bad_mode)

    def testIteration(self):
        # Test the complex interaction when mixing file-iteration and the
        # various read* methods.
        dataoffset = 16384
        filler = b'ham\n'
        assert not dataoffset % len(filler), \
            'dataoffset must be multiple of len(filler)'
        nchunks = dataoffset // len(filler)
        testlines = [
            b'spam, spam and eggs\n',
            b'eggs, spam, ham and spam\n',
            b'saussages, spam, spam and eggs\n',
            b'spam, ham, spam and eggs\n',
            b'spam, spam, spam, spam, spam, ham, spam\n',
            b'wonderful spaaaaaam.\n',
        ]
        methods = [('readline', ()), ('read', ()), ('readlines', ()),
                   ('readinto', (array('b', b' ' * 100),))]

        # Prepare the testfile
        bag = gio_pyio.open(self.file, 'wb')
        bag.write(filler * nchunks)
        bag.writelines(testlines)
        bag.close()
        # Test for appropriate errors mixing read* and iteration
        for methodname, args in methods:
            f = gio_pyio.open(self.file, 'rb', native=False)
            self.assertEqual(next(f), filler)
            meth = getattr(f, methodname)
            meth(*args)  # This simply shouldn't fail
            f.close()

        # Test to see if harmless (by accident) mixing of read* and
        # iteration still works. This depends on the size of the internal
        # iteration buffer (currently 8192,) but we can test it in a
        # flexible manner.  Each line in the bag o' ham is 4 bytes
        # ("h", "a", "m", "\n"), so 4096 lines of that should get us
        # exactly on the buffer boundary for any power-of-2 buffersize
        # between 4 and 16384 (inclusive).
        f = gio_pyio.open(self.file, 'rb', native=False)
        for _i in range(nchunks):
            next(f)
        testline = testlines.pop(0)
        try:
            line = f.readline()
        except ValueError:
            self.fail('readline() after next() with supposedly empty '
                      'iteration-buffer failed anyway')
        if line != testline:
            self.fail('readline() after next() with empty buffer '
                      'failed. Got %r, expected %r' % (line, testline))
        testline = testlines.pop(0)
        buf = array('b', b'\x00' * len(testline))
        try:
            f.readinto(buf)
        except ValueError:
            self.fail('readinto() after next() with supposedly empty '
                      'iteration-buffer failed anyway')
        line = buf.tobytes()
        if line != testline:
            self.fail('readinto() after next() with empty buffer '
                      'failed. Got %r, expected %r' % (line, testline))

        testline = testlines.pop(0)
        try:
            line = f.read(len(testline))
        except ValueError:
            self.fail('read() after next() with supposedly empty '
                      'iteration-buffer failed anyway')
        if line != testline:
            self.fail('read() after next() with empty buffer '
                      'failed. Got %r, expected %r' % (line, testline))
        try:
            lines = f.readlines()
        except ValueError:
            self.fail('readlines() after next() with supposedly empty '
                      'iteration-buffer failed anyway')
        if lines != testlines:
            self.fail('readlines() after next() with empty buffer '
                      'failed. Got %r, expected %r' % (line, testline))
        f.close()

        # Reading after iteration hit EOF shouldn't hurt either
        f = gio_pyio.open(self.file, 'rb', native=False)
        try:
            for _line in f:
                pass
            try:
                f.readline()
                f.readinto(buf)
                f.read()
                f.readlines()
            except ValueError:
                self.fail('read* failed after next() consumed file')
        finally:
            f.close()

    def testAbles(self):
        try:
            f = gio_pyio.open(self.file, 'w', native=False)
            self.assertEqual(f.readable(), False)
            self.assertEqual(f.writable(), True)
            self.assertEqual(f.seekable(), True)
            f.close()

            f = gio_pyio.open(self.file, 'r', native=False)
            self.assertEqual(f.readable(), True)
            self.assertEqual(f.writable(), False)
            self.assertEqual(f.seekable(), True)
            f.close()

            f = gio_pyio.open(self.file, 'a+', native=False)
            self.assertEqual(f.readable(), True)
            self.assertEqual(f.writable(), True)
            self.assertEqual(f.seekable(), True)
            self.assertEqual(f.isatty(), False)
            f.close()
        finally:
            pass

    def testAppend(self):
        try:
            f = gio_pyio.open(self.file, 'wb', native=False)
            f.write(b'spam')
            f.close()
            f = gio_pyio.open(self.file, 'ab', native=False)
            f.write(b'eggs')
            f.close()
            f = gio_pyio.open(self.file, 'rb', native=False)
            d = f.read()
            f.close()
            self.assertEqual(d, b'spameggs')
        finally:
            pass

    def testJSON(self):
        path = Path(Path(__file__).parent, 'example_data.json')
        file = Gio.File.new_for_path(str(path))
        f = gio_pyio.open(file, 'rb', native=False)
        data = json.load(f)
        f.close()
        assert data['glossary']['title'] == 'example glossary'

    def testGResource(self):
        parent = Path(__file__).parent
        subprocess.call(
            'glib-compile-resources' +
            ' --sourcedir=' + str(parent) +
            ' --target=' + self.file.peek_path() +
            ' ' + str(Path(parent, 'example.gresource.xml')),
            shell=True,
        )
        resource = Gio.Resource.load(self.file.peek_path())
        resource._register()

        file = Gio.File.new_for_uri('resource:///example/example_data.json')
        f = gio_pyio.open(file, 'rb')
        data = json.load(f)
        f.close()
        assert data['glossary']['title'] == 'example glossary'
